/* $Id: isaac_physicalmodel.c,v 1.12 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.12 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>
#include <float.h>
#include <cpl.h>

#include "irplib_utils.h"

#include "isaac_physicalmodel.h"
#include "isaac_pfits.h"

/*-----------------------------------------------------------------------------
   								Define
 -----------------------------------------------------------------------------*/

#define ISAAC_LR_DIR                 7.0    /* angle in degrees */
#define ISAAC_MR_DIR                31.5    /* angle in degrees */

#define ISAAC_LR_GRATING            40.0    /* grooves/mm */
#define ISAAC_MR_GRATING            210.0   /* grooves/mm */

#define ISAAC_FOCAL_LENGTH_MM       175.0   /* mm */

#define ISAAC_PIXEL_SIZE_S          18.5    /* microns */
#define ISAAC_PIXEL_SIZE_M          27.0    /* microns */

#define ISAAC_BEAM_DIFF             2.72    /* angle in degrees */
#define ISAAC_PUPIL_SIZE_MM         100.0   /* mm */

#define ISAAC_FLGTH_S1              1.75
#define ISAAC_FLGTH_S2              3.25
#define ISAAC_FLGTH_L1              1.56
#define ISAAC_FLGTH_L2              4.77
#define ISAAC_FLGTH_L3              9.88

#define ANGLE_IN_DEFAULT            0.00
#define ANGLE_OUT_DEFAULT           0.00

/*----------------------------------------------------------------------------*/
/**
 * @defgroup isaac_physicalmodel    Physical model of ISAAC
 */
/*----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
                                Function codes
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Estimate the instrument wavelength range.
  @param    filename    input file name
  @param    poly_deg    polynomial degree
  @return   1 newly allocated array of poly_deg+1 polynomical coefficients
  
  From a physical model of the instrument, find out the wavelength range
  associated to a given instrument configuration. The returned coefficients
  are such as wave = c[0] + c[1] * pix + ... + c[poly_deg] * pix^poly_deg 

  Instrument configuration informations are fetched from the FITS header.
 */
/*----------------------------------------------------------------------------*/
double * isaac_get_disprel_estimate(const char * filename, 
                                    cpl_size     poly_deg)
{
    double           * disprel = NULL;
    cpl_propertylist * plist = cpl_propertylist_load(filename, 0);
    const char       * objective;
    const char       * resolution;
    double             wl_c;
    int                npix;
    int                order;

    cpl_matrix       * x_val = NULL;
    cpl_vector       * y_val = NULL;
    cpl_polynomial   * poly_fit = NULL;
    double           * coefs = NULL;


    bug_if(0);

    /* A polynomial fit is computed  */
    skip_if_lt (poly_deg, 0, "polynomial degree");

    /* Get the central wavelength */
    wl_c = isaac_pfits_get_wlen(plist);
    error_if(cpl_error_get_code(), cpl_error_get_code(), 
             "cannot get central wlen from [%s]", filename);

    wl_c *= 10000.0; /* microns to A */

    /* Get the objective */
    objective = isaac_pfits_get_objective(plist);
    error_if(cpl_error_get_code(), cpl_error_get_code(),
             "cannot get objective from [%s]", filename);

    /* Get the grating name */
    resolution = isaac_pfits_get_resolution(plist);
    error_if(cpl_error_get_code(), cpl_error_get_code(),
             "cannot get resolution from [%s]", filename);

    /* Display configuration */
    cpl_msg_info(cpl_func, "configuration for ISAAC physical model: ");
    cpl_msg_info(cpl_func, "resolution: %s", resolution);
    cpl_msg_info(cpl_func, "lambda_c   [Angstrom]: %g", wl_c);
    cpl_msg_info(cpl_func, "objective    : %s", objective);

    /* Get naxis1 */
    npix = isaac_pfits_get_naxis1(plist);
    error_if(cpl_error_get_code(), cpl_error_get_code(), 
             "cannot get NAXIS1 from [%s]", filename);

    skip_if_lt(npix, 2, "NAXIS1");

    /* Set the order - as in isaac_physical_model() */
    if ( wl_c >= 8900 && wl_c < 9900 ) {
        order = 6;
    } else if ( wl_c >= 9900 && wl_c < 11000 ) {
        order = 5;
    } else if ( wl_c >= 11000 && wl_c < 14000 ) {
        order = 4;
    } else if ( wl_c >= 14000 && wl_c < 18500 ) {
        order = 3;
    } else if ( wl_c >= 18500 && wl_c < 25000 ) {
        order = 2;
    } else {
        order = 1;
    }

    disprel = (double*)cpl_malloc((poly_deg+1) * sizeof(double));

    if (poly_deg == 3 && resolution[0] == 'M' && 8900 <= wl_c && wl_c < 25000) {

        /* This method and the constants are provided by C. Lidman */

        const double  a = -1.218717e-7;
        const double  b =  0.003395204;
        const double  c =  1337.455;
        const double  d = -1.617833e-4;
        const double  e =  3.132269;
        const double  f = -2.496095;

        const double c1 = wl_c;
        const double c2 = (( a * (order*c1) + b) * (order*c1) + c) / order;
        const double c3 = ( d * (order*c1) + e ) / order;
        const double c4 = f / order;
        double       k0 = -(npix+1) / (double)(npix-1);
        double       k1 =          2 / (double)(npix-1);

        double     * p  = disprel;

        /* The polynomial in reduced coordinates, -1 <= z <= 1 */
        p[0] = c1 - c3;
        p[1] = c2 - 3*c4;
        p[2] = 2*c3;
        p[3] = 4*c4;
        cpl_msg_debug(cpl_func, 
                "Reduced polynomial(%d:%d): %g + %g * z + %g * z^2 "
                "+ %g * z^3\n", order, npix, p[0], p[1], p[2], p[3]);

        /* The polynomial in pixel coordinates, 1 <= pix <= npix */
        p[0] += k0 * ( p[1] + k0 * ( p[2] + k0 * p[3]));

        p[1] += k0 * ( 2*p[2] + k0 * 3*p[3]);
        p[1] *= k1;

        p[2] += k0 * 3*p[3];
        p[2] *= k1*k1;

        p[3] *= k1*k1*k1;

    } else {
        double          * px_val;
        const cpl_boolean sampsym = CPL_TRUE;
        cpl_size          i;

        /* coefs an array of npix doubles with the wavelength for each pix. */
        coefs = isaac_physical_model(wl_c, objective, resolution, npix);
        error_if (coefs == NULL, cpl_error_get_code(),
                  "cannot compute the physical model");

        y_val = cpl_vector_wrap(npix, coefs);

        x_val = cpl_matrix_new(1, npix);
        px_val = cpl_matrix_get_data(x_val);

        for (i = 0; i < npix; i++) {
            px_val[i] = (double)(i+1);
        }

        /* Apply the fit */
        poly_fit = cpl_polynomial_new(1);
        skip_if(cpl_polynomial_fit(poly_fit, x_val, &sampsym, y_val, NULL,
                                   CPL_FALSE, NULL, &poly_deg));
        /* Extract coefs (FIXME: use cpl_polynomial...) */
        for (i = 0; i <= poly_deg; i++) {
            disprel[i] = cpl_polynomial_get_coeff(poly_fit, &i); 
        }
    }

    end_skip;

    if (cpl_error_get_code()) {
        cpl_free(disprel);
        disprel = NULL;
    }

    cpl_matrix_delete(x_val);
    cpl_polynomial_delete(poly_fit);
    (void)cpl_vector_unwrap(y_val);
    cpl_free(coefs);

    cpl_propertylist_delete(plist);

    return disprel;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    ISAAC physical model
  @param    lambda_c    central wavelength
  @param    objective   slit used
  @param    resolution  resolution (medium or low)
  @param    nbpix       number of pixels to calibrate
  @return   wavelengths array (index i <-> pixel nb i+1)
    
  This module determines the dispersion relation of ISAAC for the different 
  configurations of objectives, gratings, detectors, and central wavelengths.

  Assumed Optical configuration:
   
    Focal lens of objectives for short- and long-wavelength objective (at 77 K):
        S1 = f/1.75, S2 = f/3.25, L1 = f/1.56, L2 = f/4.77,  L3 = f/9.98
    Pupil size: 100 mm
    Pixel size: 18.5 microns (SW), 27 microns (LW)
    Gratings:   low resolution, 40 gr/mm, entering at about 5 degrees
                medium resolution, 210 gr/mm, entering at about 23 degrees
    Beam difference: 2.72 degrees
 */
/*----------------------------------------------------------------------------*/
double * isaac_physical_model(double       lambda_c,
                              const char * objective,
                              const char * resolution,
                              int          nbpix)
{
    double          beam_diff, focal_length, pixel_size, pupil; 
    double          gr;
    double          isaacLRGrating, isaacMRGrating;
    int             order;
    double      *   disp;
    double          x;
    double          angle_in, angle_out;
    double          a, b, det;
    double          loc_pi;
    int             i;

    /* Initialize */
    isaacLRGrating  = ISAAC_LR_GRATING;
    isaacMRGrating  = ISAAC_MR_GRATING;
    focal_length    = ISAAC_FOCAL_LENGTH_MM;
    pixel_size      = ISAAC_PIXEL_SIZE_S;
    beam_diff       = ISAAC_BEAM_DIFF;
    pupil           = ISAAC_PUPIL_SIZE_MM;
    loc_pi          = 3.1415926535897932384626433;

    /* Check objective */
    if (objective[0] == 'S') {
        if (objective[1] == '1') focal_length = pupil*ISAAC_FLGTH_S1;
        if (objective[1] == '2') focal_length = pupil*ISAAC_FLGTH_S2;
        pixel_size = ISAAC_PIXEL_SIZE_S;
    } else if (objective[0] == 'L') {
        if (objective[1] == '1') focal_length = pupil*ISAAC_FLGTH_L1;
        if (objective[1] == '2') focal_length = pupil*ISAAC_FLGTH_L2;
        if (objective[1] == '3') focal_length = pupil*ISAAC_FLGTH_L3;
        pixel_size = ISAAC_PIXEL_SIZE_M;
    } else {
        cpl_msg_error(cpl_func, "unsupported objective");
        return NULL;
    }

    /* Convert mm to m */
    focal_length *= 1e-3; 

    /* Display configuration */
    cpl_msg_info(cpl_func, "configuration for ISAAC physical model: ");
    if (resolution[0]=='L') cpl_msg_info(cpl_func, "low resolution");
    if (resolution[0]=='M') cpl_msg_info(cpl_func, "medium resolution");
    cpl_msg_info(cpl_func, "lambda_c     : %g", lambda_c);
    cpl_msg_info(cpl_func, "objective    : %s", objective);
    cpl_msg_info(cpl_func, "focal length : %g", focal_length);

    /* Set the grating */
    if (resolution[0] == 'L') { gr = isaacLRGrating;}
	else if (resolution[0] == 'M') { gr = isaacMRGrating;}
	else {    
		cpl_msg_error(cpl_func, "wrong grating! %c", resolution[0]);
        return NULL;
    }

    /* Convert gr to grove/nm */
    gr *= 1e-6;  
    /* Convert beam_diff in radians. */
    beam_diff *= loc_pi / 180.0; 
    /* Convert lambda_c : A->nm */
    lambda_c /= 10;

    /* Set the order */
    if ( lambda_c >= 890 && lambda_c < 8000) {
        if ( lambda_c >= 890 && lambda_c < 990 ) {
            order = 6;
        } else if ( lambda_c >= 990 && lambda_c < 1100 ) {
            order = 5;
        } else if ( lambda_c >= 1100 && lambda_c < 1400 ) {
            order = 4;
        } else if ( lambda_c >= 1400 && lambda_c < 1850 ) {
            order = 3;
        } else if ( lambda_c >= 1850 && lambda_c < 2500 ) {
            order = 2;
        } else if ( lambda_c >= 2500 && lambda_c < 5500 ) {
            order = 1;
        } else {
            order = 1;
        }
    } else {
        order = (int)(0.5 + (sin(ANGLE_IN_DEFAULT) + sin(ANGLE_OUT_DEFAULT)) /
                (lambda_c * gr));
        if (order < 1) {
            cpl_msg_error(cpl_func, "wrong order! %d", order);
            return NULL;
        }
    }
    cpl_msg_info(cpl_func, "order    : %d", order);

    /* The following is the solution of the set of equations : */
    /* (1)  sin(angle_in) + sin(angle_out) = order*gr*lambda_c  */
    /* (2)  angle_out - angle_in = beam_diff */
    det = 2 * sin(beam_diff);
    a   = order * gr * lambda_c * sin(beam_diff);
    b   = sqrt(order * gr * lambda_c * order * gr * lambda_c *
        (sin(beam_diff) * sin(beam_diff) + 2 * cos(beam_diff) - 2)
        + 2 * sin(beam_diff) * sin(beam_diff) * (1-cos(beam_diff)));
    angle_in = asin((a-b)/det);
    angle_out = angle_in + beam_diff;

    disp = cpl_malloc(nbpix * sizeof(double));

    for (i=0; i<nbpix; i++) {
        /* x in meters */
        x = (double)(i-(nbpix/2.)) * pixel_size * 1e-6; 
        disp[i]=(sin(angle_in)+sin(angle_out+atan(x/focal_length)))/(order*gr);

        /* Convert disprel from nm to angstroms */
        disp[i] *= 10;
    }

    return disp;
}

/**@}*/

