/* $Id: isaac_img_illum.c,v 1.39 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.39 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <cpl.h>

#include "irplib_plugin.h"
#include "irplib_utils.h"
#include "irplib_calib.h"
#include "irplib_strehl.h"

#include "isaac_utils.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "isaac_img_illum"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static cpl_image * isaac_img_illum_reduce(const cpl_frameset *, const char *,
                                          const char *, const char *,
                                          cpl_table **);
static cpl_bivector * isaac_img_illum_find_pos(const cpl_imagelist *,
                                               const cpl_bivector *);
static cpl_vector * isaac_img_illum_phot(const cpl_imagelist *,
                                         const cpl_bivector *, int *);
static cpl_error_code isaac_img_illum_save(cpl_frameset *,
                                           const cpl_parameterlist *,
                                           const cpl_image *, const cpl_table *);

cpl_recipe_define(isaac_img_illum, ISAAC_BINARY_VERSION,
                  "Lars Lundin", PACKAGE_BUGREPORT, "2008", 
                  "ISAAC imaging illumination frame recipe",
                  RECIPE_STRING " -- ISAAC imaging illumination frame recipe.\n"
                  "The files listed in the Set Of Frames (sof-file) " 
                  "must be tagged:\n"
                  "raw-file.fits "  ISAAC_IMG_ILLUM_RAW" or\n"
                  "flat-file.fits " ISAAC_CALIB_FLAT" or\n"
                  "bpm-file.fits "  ISAAC_CALIB_BPM" or\n"
                  "dark-file.fits " ISAAC_CALIB_DARK"\n");

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int         s_hx;
    int         s_hy;
    double      star_r;
    double      bg_r1;
    double      bg_r2;
    int         pos_x;
    int         pos_y;

    /* Outputs */
    cpl_size    poly_deg;
    double      fit_error;
    double      norm;
    double      illum1;
    double      illumx;
    double      illumy;
    double      illumxy;
    double      illumxx;
    double      illumyy;
} isaac_img_illum_config;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Fill the recipe parameterlist
  @param    self  The parameterlist
  @return   0 iff everything is ok

 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code isaac_img_illum_fill_parameterlist(cpl_parameterlist * self)
{
    const char * context = PACKAGE "." RECIPE_STRING;
    cpl_error_code err;

    cpl_ensure_code(self, CPL_ERROR_NULL_INPUT);

    /* Fill the parameters list */

    /* --star_r */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING,
                                          "star_r", 10.0, NULL, context,
                                          "The star radius");
    cpl_ensure_code(!err, err);

    /* --bg_r1 */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING,
                                          "bg_r1", 12.0, NULL, context,
                                          "The internal background radius");
    cpl_ensure_code(!err, err);

    /* --bg_r2 */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING,
                                          "bg_r2", 30.0, NULL, context,
                                          "The external background radius");
    cpl_ensure_code(!err, err);

    /* --s_hx */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "s_hx", 50, NULL, context,
                                       "Half size of the seach box in x");
    cpl_ensure_code(!err, err);

    /* --s_hy */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "s_hy", 50, NULL, context,
                                       "Half size of the seach box in y");
    cpl_ensure_code(!err, err);

    /* --pos_x */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "pos_x", -1, NULL, context,
                                       "The star position in x");
    cpl_ensure_code(!err, err);

    /* --pos_y */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "pos_y", -1, NULL, context,
                                       "The star position in y");
    cpl_ensure_code(!err, err);

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Get the command line options and execute the data reduction
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_illum(cpl_frameset            * framelist,
                           const cpl_parameterlist * parlist)
{
    cpl_frameset * rawframes = NULL;
    cpl_image    * ima       = NULL;
    cpl_table    * flux      = NULL;
    const char   * bpm;
    const char   * flat;
    const char   * dark;


    /* Retrieve input parameters */
    isaac_img_illum_config.s_hx
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING, "s_hx");

    isaac_img_illum_config.s_hy
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING, "s_hy");

    isaac_img_illum_config.star_r
        = irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING,
                                          "star_r");
    isaac_img_illum_config.bg_r1
        = irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING,
                                          "bg_r1");
    isaac_img_illum_config.bg_r2
        = irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING,
                                          "bg_r2");
    isaac_img_illum_config.pos_x
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "pos_x");
    isaac_img_illum_config.pos_y
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "pos_y");

    /* Identify the RAW and CALIB frames in the input frameset */
    skip_if (isaac_dfs_set_groups(framelist));

    /* Retrieve calibration data */
    flat = isaac_extract_filename(framelist, ISAAC_CALIB_FLAT);
    dark = isaac_extract_filename(framelist, ISAAC_CALIB_DARK);
    bpm  = isaac_extract_filename(framelist, ISAAC_CALIB_BPM);

    /* Retrieve raw frames */
    rawframes = isaac_extract_frameset(framelist, ISAAC_IMG_ILLUM_RAW);
    error_if(rawframes == NULL, CPL_ERROR_DATA_NOT_FOUND,
             "Cannot find any raw frame in the input");
    
    /* Apply the data reduction */
    cpl_msg_info(cpl_func, "Apply the reduction");
    ima = isaac_img_illum_reduce(rawframes, flat, dark, bpm, &flux);
    error_if (ima == NULL, cpl_error_get_code(), "Cannot reduce the data");

    /* Save the products */
    cpl_msg_info(cpl_func, "Saving the products");
    skip_if (isaac_img_illum_save(framelist, parlist, ima, flux));

    end_skip;

    cpl_image_delete(ima);
    cpl_table_delete(flux);
    cpl_frameset_delete(rawframes);

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    raw         the raw frames
  @param    flat        the flat field or NULL
  @param    dark        the dark or NULL
  @param    bpm         the bad pixels map or NULL
  @param    fl          the returned flux table
  @return   the illumination image or NULL on error
 */
/*----------------------------------------------------------------------------*/
static cpl_image * isaac_img_illum_reduce(const cpl_frameset * raw,
                                          const char         * flat,
                                          const char         * dark,
                                          const char         * bpm,
                                          cpl_table         ** fl)
{
    cpl_bivector    *   offsets;
    cpl_imagelist   *   in;
    cpl_bivector    *   positions;
    cpl_matrix      *   purged_pos;
    double          *   positions_x,
                    *   positions_y,
                    *   purged_pos_x,
                    *   purged_pos_y;
    cpl_vector      *   flux;
    cpl_vector      *   purged_flux;
    cpl_polynomial  *   poly;
    cpl_image       *   poly_ima;
    int                 nval = 0; /* uninit warning */
    int                 nx, ny;
    int                 nflux;
    cpl_size            power[2];
    int                 i, j;
    
    /* Initialise */
    *fl = NULL;

    /* Get the offsets from the raw frames */
    cpl_msg_info(cpl_func, "Get the offsets from the files headers");
    cpl_msg_indent_more();
    if ((offsets = isaac_get_offsets(raw)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot get the offsets");
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_msg_indent_less();

    /* Load the input data */
    cpl_msg_info(cpl_func, "Load the input data");
    cpl_msg_indent_more();
    if ((in = cpl_imagelist_load_frameset(raw, CPL_TYPE_FLOAT, 1, 0)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot load input data");
        cpl_bivector_delete(offsets);
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_msg_indent_less();

    /* Apply the calibrations */
    if (flat || dark || bpm) {
        cpl_msg_info(cpl_func, "Apply the calibrations");
        cpl_msg_indent_more();
        if (irplib_flat_dark_bpm_calib(in, flat, dark, bpm) == -1) {
            cpl_msg_error(cpl_func, "Cannot calibrate the data");
            cpl_imagelist_delete(in);
            cpl_bivector_delete(offsets);
            cpl_msg_indent_less();
            return NULL;
        }
        cpl_msg_indent_less();
    }

    /* Find the positions */
    cpl_msg_info(cpl_func, "Find the positions");
    cpl_msg_indent_more();
    if ((positions=isaac_img_illum_find_pos(in, offsets)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot find the positions");
        cpl_imagelist_delete(in);
        cpl_bivector_delete(offsets);
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_bivector_delete(offsets);
    cpl_msg_indent_less();

    /* Compute the photometry */
    cpl_msg_info(cpl_func, "Compute the photometry");
    cpl_msg_indent_more();
    if ((flux=isaac_img_illum_phot(in, positions, &nval)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot find the positions");
        cpl_imagelist_delete(in);
        cpl_bivector_delete(positions);
        cpl_msg_indent_less();
        return NULL;
    }
    nflux = cpl_vector_get_size(flux);
    nx = cpl_image_get_size_x(cpl_imagelist_get(in, 0));
    ny = cpl_image_get_size_y(cpl_imagelist_get(in, 0));
    cpl_imagelist_delete(in);
    cpl_msg_indent_less();
    if (nval < 1) {
        cpl_msg_error(cpl_func, "No flux computed");
        cpl_bivector_delete(positions);
        cpl_vector_delete(flux);
        return NULL;
    }
    
    /* Purge positions */
    purged_pos = cpl_matrix_new(2, nval);
    purged_pos_x = cpl_matrix_get_data(purged_pos);
    purged_pos_y = purged_pos_x + nval;
    positions_x = cpl_bivector_get_x_data(positions);
    positions_y = cpl_bivector_get_y_data(positions);
    purged_flux = cpl_vector_new(nval);
    j = 0;
    for (i=0; i < nflux; i++) {
        if (fabs(cpl_vector_get(flux, i)) > 0) {
            purged_pos_x[j] = positions_x[i];
            purged_pos_y[j] = positions_y[i];
            cpl_vector_set(purged_flux, j, cpl_vector_get(flux, i));
            j++;
        }
    }

    if (j != nval) {
        cpl_msg_warning(cpl_func, "Fixing size inconsistency: %d <=> %d",
                        j, nval);
        cpl_matrix_set_size(purged_pos, 2, j);
        cpl_vector_set_size(purged_flux, j);
    }

    /* Create the flux table */
    *fl = cpl_table_new(cpl_vector_get_size(flux));
    cpl_table_new_column(*fl, "POSX", CPL_TYPE_DOUBLE);
    cpl_table_new_column(*fl, "POSY", CPL_TYPE_DOUBLE);
    cpl_table_new_column(*fl, "FLUX", CPL_TYPE_DOUBLE);
    for (i=0; i<cpl_vector_get_size(flux); i++) {
        cpl_table_set_double(*fl, "POSX", i, positions_x[i]);
        cpl_table_set_double(*fl, "POSY", i, positions_y[i]);
        cpl_table_set_double(*fl, "FLUX", i, cpl_vector_get(flux, i));
    }
    cpl_bivector_delete(positions);
    cpl_vector_delete(flux);
    
    /* Degree of the polynomial */
    if (nval >= 6)
        isaac_img_illum_config.poly_deg = 2;
    else if (nval >= 3) isaac_img_illum_config.poly_deg = 1;
    else isaac_img_illum_config.poly_deg = 0;

    cpl_msg_info(cpl_func, "Fitting a %d-degree 2D-polynomial using %d of %d "
                 "fluxes", (int)isaac_img_illum_config.poly_deg, nval, nflux);

    /* Fit the polynomial */
    poly = cpl_polynomial_new(2);
    if (cpl_polynomial_fit(poly, purged_pos, NULL, purged_flux, NULL, CPL_FALSE,
                           NULL, &(isaac_img_illum_config.poly_deg))) {
        cpl_msg_error(cpl_func, "Cannot fit the polynomial");
        cpl_matrix_delete(purged_pos);
        cpl_vector_delete(purged_flux);
        cpl_polynomial_delete(poly);
        cpl_table_delete(*fl);
        *fl = NULL;
        return NULL;
    }

    cpl_vector_fill_polynomial_fit_residual(purged_flux, purged_flux, NULL, poly,
                                            purged_pos, NULL);

    cpl_matrix_delete(purged_pos);

    isaac_img_illum_config.fit_error
        = cpl_vector_product(purged_flux, purged_flux)
        / cpl_vector_get_size(purged_flux);

    if (cpl_msg_get_level() <= CPL_MSG_DEBUG)
        cpl_vector_dump(purged_flux, stdout);

    cpl_vector_delete(purged_flux);
   
    /* Fill the coeffs */
    power[0] = 0; power[1] = 0;
    isaac_img_illum_config.illum1 = cpl_polynomial_get_coeff(poly, power);
    if (isaac_img_illum_config.poly_deg >= 1) {
        power[0] = 1; power[1] = 0;
        isaac_img_illum_config.illumx = cpl_polynomial_get_coeff(poly, power);
        power[0] = 0; power[1] = 1;
        isaac_img_illum_config.illumy = cpl_polynomial_get_coeff(poly, power);
    }
    if (isaac_img_illum_config.poly_deg >= 2) {
        power[0] = 1; power[1] = 1;
        isaac_img_illum_config.illumxy = cpl_polynomial_get_coeff(poly, power);
        power[0] = 2; power[1] = 0;
        isaac_img_illum_config.illumxx = cpl_polynomial_get_coeff(poly, power);
        power[0] = 0; power[1] = 2;
        isaac_img_illum_config.illumyy = cpl_polynomial_get_coeff(poly, power);
    }
    cpl_msg_info(cpl_func, "P(X,Y)=a0+a1*X+a2*Y+a3*X*Y+a4*X^2+a5*Y^2");
    cpl_msg_info(cpl_func, "a0 = %g", isaac_img_illum_config.illum1);
    cpl_msg_info(cpl_func, "a1 = %g", isaac_img_illum_config.illumx);
    cpl_msg_info(cpl_func, "a2 = %g", isaac_img_illum_config.illumy);
    cpl_msg_info(cpl_func, "a3 = %g", isaac_img_illum_config.illumxy);
    cpl_msg_info(cpl_func, "a4 = %g", isaac_img_illum_config.illumxx);
    cpl_msg_info(cpl_func, "a5 = %g", isaac_img_illum_config.illumyy);
    cpl_msg_info(cpl_func, "error = %g", isaac_img_illum_config.fit_error);
    
    /* Create the polynomial image */
    poly_ima = cpl_image_new(nx, ny, CPL_TYPE_FLOAT);
    cpl_image_fill_polynomial(poly_ima, poly, 1.0, 1.0, 1.0, 1.0);
    cpl_polynomial_delete(poly);

    /* Normalise the image */
    isaac_img_illum_config.norm = cpl_image_get_mean(poly_ima);
    cpl_image_divide_scalar(poly_ima, isaac_img_illum_config.norm);

    return poly_ima;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the photometry of the bright star in the images.
  @param    ilist       the images list
  @param    positions   the star positions
  @param    nvalid      number of valid points
  @return   the flux in a vector or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * isaac_img_illum_phot(const cpl_imagelist * ilist,
                                         const cpl_bivector  * positions,
                                         int                 * nvalid)
{
    const int      ni          = cpl_imagelist_get_size(ilist);
    /* Create the flux vector */
    cpl_vector   * self        = cpl_vector_new(ni);
    const double * positions_x = cpl_bivector_get_x_data_const(positions);
    const double * positions_y = cpl_bivector_get_y_data_const(positions);
    int            nok         = 0;
    int            i;


    bug_if(ilist     == NULL);
    bug_if(positions == NULL);
    bug_if(nvalid    == NULL);

    /* Loop on the frames */
    for (i=0; i < ni; i++) {
        const cpl_image * cur_ima = cpl_imagelist_get_const(ilist, i);
        const double bg
            = irplib_strehl_ring_background(cur_ima, positions_x[i],
                                            positions_y[i],
                                            isaac_img_illum_config.bg_r1,
                                            isaac_img_illum_config.bg_r2,
                                            IRPLIB_BG_METHOD_MEDIAN);
        const double fl
            = irplib_strehl_disk_flux(cur_ima, positions_x[i], positions_y[i],
                                      isaac_img_illum_config.star_r, bg);
        cpl_vector_set(self, i, fl); 
        if (fabs(fl)>1e-3) nok++; 
        cpl_msg_info(cpl_func, "Image %2d has flux at (%g,%g): %g", i+1,
                     positions_x[i], positions_y[i], fl);
    }

    *nvalid = nok;

    end_skip;

    if (cpl_error_get_code()) {
        cpl_vector_delete(self);
        self = NULL;
    }

    return self;
}
 
/*----------------------------------------------------------------------------*/
/**
  @brief    Get the positions of the bright star in the images.
  @param    ilist       the images list
  @param    offsets     the offsets between frames
  @return   the positions or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_bivector * isaac_img_illum_find_pos(const cpl_imagelist * ilist,
                                               const cpl_bivector  * offsets)
{
    const int      ni = cpl_imagelist_get_size(ilist);
    /* Create the positions bivector */
    cpl_bivector * self = cpl_bivector_new(ni);
    double       * self_x = cpl_bivector_get_x_data(self);
    double       * self_y = cpl_bivector_get_y_data(self);
    const double * offsets_x = cpl_bivector_get_x_data_const(offsets);
    const double * offsets_y = cpl_bivector_get_y_data_const(offsets);
    cpl_image    * tmp_ima   = NULL;
    cpl_image    * filtered_ima = NULL;
    cpl_mask     * kernel = cpl_mask_new(3, 3);
    int            i;

    bug_if(ilist   == NULL);
    bug_if(offsets == NULL);
    
    /* Create the kernel */
    bug_if(cpl_mask_not(kernel));
    
    /* Loop on all the images */
    for (i=0; i < ni; i++) {
        const cpl_image * cur_ima = cpl_imagelist_get_const(ilist, i);
        const int         nx = cpl_image_get_size_x(cur_ima);
        const int         ny = cpl_image_get_size_y(cur_ima);
        int               llx, lly, urx, ury;
        cpl_size          posx, posy;

        /* Define zone */
        if (isaac_img_illum_config.pos_x > 0 &&
            isaac_img_illum_config.pos_y > 0) {
            llx = isaac_img_illum_config.pos_x + offsets_x[i] - 
                isaac_img_illum_config.s_hx;
            urx = isaac_img_illum_config.pos_x + offsets_x[i] + 
                isaac_img_illum_config.s_hx;
            lly = isaac_img_illum_config.pos_y + offsets_y[i] - 
                isaac_img_illum_config.s_hy;
            ury = isaac_img_illum_config.pos_y + offsets_y[i] + 
                isaac_img_illum_config.s_hy;
        } else {
            llx = nx/2 + offsets_x[i] - isaac_img_illum_config.s_hx;
            urx = nx/2 + offsets_x[i] + isaac_img_illum_config.s_hx;
            lly = ny/2 + offsets_y[i] - isaac_img_illum_config.s_hy;
            ury = ny/2 + offsets_y[i] + isaac_img_illum_config.s_hy;
        }

        /* Extract */
        cpl_image_delete(tmp_ima);
        tmp_ima = cpl_image_extract(cur_ima, llx, lly, urx, ury);
        any_if ("Bad star zone. llx=%d, urx=%d, nx=%d. "
                "lly=%d, ury=%d, ny=%d", llx, urx, nx, lly, ury, ny);

        cpl_image_delete(filtered_ima);
        filtered_ima = cpl_image_new(urx-llx+1, ury-lly+1,
                                     cpl_image_get_type(tmp_ima));
        cpl_image_filter_mask(filtered_ima, tmp_ima, kernel, CPL_FILTER_MEDIAN,
                              CPL_BORDER_FILTER);

        /* Find the max */
        bug_if(cpl_image_get_maxpos(filtered_ima, &posx, &posy));
        self_x[i] = llx + posx;
        self_y[i] = lly + posy;
        cpl_msg_info(cpl_func, "image %2d has star at: %g, %g", i+1,
                     self_x[i], self_y[i]);
    }

    end_skip;

    cpl_image_delete(tmp_ima);
    cpl_image_delete(filtered_ima);
    cpl_mask_delete(kernel);

    if (cpl_error_get_code()) {
        cpl_bivector_delete(self);
        self = NULL;
    }

    return self;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the different products of the recipe
  @param    ima         the fitted image
  @param    flux        the flux table (optional)
  @param    parlist     the input list of parameters
  @param    set         the input frame set
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code isaac_img_illum_save(cpl_frameset            * set,
                                           const cpl_parameterlist * parlist,
                                           const cpl_image         * ima,
                                           const cpl_table         * flux)
{
    /* Get FITS header from reference file */
    const char * filename     = cpl_frame_get_filename
        (irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW));
    cpl_propertylist * plist  = NULL;
    cpl_propertylist * qclist = cpl_propertylist_new();
    const char       * sval;

    skip_if(0);

    plist = cpl_propertylist_load(filename, 0);
    skip_if(plist == NULL);

    sval = isaac_pfits_get_filter(plist);
    if (cpl_error_get_code()) cpl_error_reset();
    else cpl_propertylist_append_string(qclist, "ESO QC FILTER OBS", sval);

    cpl_propertylist_append_double(qclist, "ESO QC ILLUM1",
            isaac_img_illum_config.illum1);
    cpl_propertylist_append_double(qclist, "ESO QC ILLUMX", 
            isaac_img_illum_config.illumx);
    cpl_propertylist_append_double(qclist, "ESO QC ILLUMY", 
            isaac_img_illum_config.illumy);
    cpl_propertylist_append_double(qclist, "ESO QC ILLUMXY",
            isaac_img_illum_config.illumxy);
    cpl_propertylist_append_double(qclist, "ESO QC ILLUMXX", 
            isaac_img_illum_config.illumxx);
    cpl_propertylist_append_double(qclist, "ESO QC ILLUMYY",
            isaac_img_illum_config.illumyy);
    cpl_propertylist_append_double(qclist, "ESO QC FITERROR",
            isaac_img_illum_config.fit_error);
    cpl_propertylist_append_double(qclist, "ESO QC FITNORM",
            isaac_img_illum_config.norm);

    bug_if(0);

    /* Get the keywords for the paf file */
    (void)cpl_propertylist_erase_regexp(plist,
                                        "^(ARCFILE|MJD-OBS|ESO TPL ID|DATE-OBS"
                                        "|ESO DET DIT|ESO DET NDIT"
                                        "|ESO DET NCORRS|ESO DET MODE NAME"
                                        "|ESO OCS SELECT-ARM)$", 1);
    
    /* Write the image */
    skip_if(irplib_dfs_save_image(set, parlist, set, ima, CPL_BPP_IEEE_FLOAT,
                                  RECIPE_STRING, ISAAC_IMG_ILLUM_RES,
                                  qclist, NULL, PACKAGE "/" PACKAGE_VERSION,
                                  RECIPE_STRING CPL_DFS_FITS));

    if (flux != NULL) {
        /* Write the table */
        skip_if(irplib_dfs_save_table(set, parlist, set, flux, NULL,
                                      RECIPE_STRING, ISAAC_IMG_ILLUM_TAB,
                                      qclist, NULL, PACKAGE "/" PACKAGE_VERSION,
                                      RECIPE_STRING "_flux" CPL_DFS_FITS));
    }
    
    /* Copy the QC in paflist */
    bug_if(cpl_propertylist_append(plist, qclist));
    cpl_propertylist_empty(qclist);

    /* PRO.CATG */
    cpl_propertylist_update_string(plist, CPL_DFS_PRO_CATG,
                                   ISAAC_IMG_ILLUM_RES);

    /* Save the PAF file */
    skip_if(cpl_dfs_save_paf("ISAAC", RECIPE_STRING, plist,
                             RECIPE_STRING CPL_DFS_PAF));

    end_skip;

    cpl_propertylist_delete(plist);
    cpl_propertylist_delete(qclist);

    return cpl_error_get_code();
}
