/* $Id: isaac_spc_flat.c,v 1.40 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.40 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>
#include <cpl.h>

#include "irplib_utils.h"

#include "isaac_utils.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"

/*-----------------------------------------------------------------------------
                                Defines
 -----------------------------------------------------------------------------*/

#define MEDIAN_XSIZE        200
#define MEDIAN_YSIZE        200

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int isaac_spc_flat_create(cpl_plugin *);
static int isaac_spc_flat_exec(cpl_plugin *);
static int isaac_spc_flat_destroy(cpl_plugin *);
static int isaac_spc_flat(cpl_parameterlist *, cpl_frameset *);
static cpl_image * isaac_spc_flat_reduce(cpl_frameset *);
static cpl_imagelist * isaac_spc_flat_diffs(cpl_frameset *); 
static cpl_image * isaac_spc_flat_divide_fit(cpl_image *, int, int, int);
static int isaac_spc_flat_save(cpl_image *, int, cpl_frameset *, 
        cpl_parameterlist *, cpl_frameset *);
static int isaac_spc_flat_compare(const cpl_frame *, const cpl_frame *); 

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    double      low_thresh;
    double      high_thresh;
    int         fit_order;
    int         fit_size;
    int         offset;
    int         llx;
    int         lly;
    int         urx;
    int         ury;
    /* Outputs */
    double      med_stdev;
    double      med_avg;
} isaac_spc_flat_config;

static char isaac_spc_flat_description[] = 
"isaac_spc_flat -- ISAAC spectro flat-field creation.\n"
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"raw-file.fits "ISAAC_SPC_FLAT_RAW"\n";

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe));
    cpl_plugin  *   plugin = &recipe->interface;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    ISAAC_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "isaac_spc_flat",
                    "Spectro flat recipe",
                    isaac_spc_flat_description,
                    "Lars Lundin",
                    PACKAGE_BUGREPORT,
                    isaac_get_license(),
                    isaac_spc_flat_create,
                    isaac_spc_flat_exec,
                    isaac_spc_flat_destroy);

    cpl_pluginlist_append(list, plugin);
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_flat_create(cpl_plugin * plugin)
{
    cpl_recipe      * recipe;
    cpl_parameter   * p;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();

    /* Fill the parameters list */
    /* --thresholds */
    p = cpl_parameter_new_value("isaac.isaac_spc_flat.thresholds", 
            CPL_TYPE_STRING, "Low and high thresholds", "isaac.isaac_spc_flat",
            "0.01,3.0");
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "thresholds");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --fit_order */
    p = cpl_parameter_new_value("isaac.isaac_spc_flat.fit_order", CPL_TYPE_INT,
            "Order for the fit", "isaac.isaac_spc_flat", 3);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "fit_order");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --fit_size */
    p = cpl_parameter_new_value("isaac.isaac_spc_flat.fit_size", CPL_TYPE_INT,
            "Size for the fit", "isaac.isaac_spc_flat", 200);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "fit_size");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --offset */
    p = cpl_parameter_new_value("isaac.isaac_spc_flat.offset", CPL_TYPE_INT,
            "Offset", "isaac.isaac_spc_flat", 40);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "offset");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --zone */
    p = cpl_parameter_new_value("isaac.isaac_spc_flat.zone", 
            CPL_TYPE_STRING, "Zone to consider", "isaac.isaac_spc_flat",
            "256,256,768,768");
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "zone");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* Return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_flat_exec(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    return isaac_spc_flat(recipe->parameters, recipe->frames);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_flat_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    cpl_parameterlist_delete(recipe->parameters);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_flat(
        cpl_parameterlist   *   parlist, 
        cpl_frameset        *   framelist)
{
    const char      *   sval;
    cpl_parameter   *   par;
    cpl_size        *   labels;
    cpl_size            nlabels;
    cpl_frameset    *   flatframes;
    cpl_frameset    *   flat_one;
    cpl_image       *   spflat;
    int                 i;
    cpl_boolean         did_reduce = CPL_FALSE;
    
    /* Initialise */
    par = NULL;
    isaac_spc_flat_config.med_stdev = 0.0;
    isaac_spc_flat_config.med_avg = 0.0;

    /* Retrieve input parameters */
    /* --thresholds */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_spc_flat.thresholds"); 
    sval = cpl_parameter_get_string(par);
    if (sscanf(sval, "%lg,%lg",
                    &isaac_spc_flat_config.low_thresh,
                    &isaac_spc_flat_config.high_thresh)!=2) {
        return CPL_ERROR_UNSPECIFIED;
    }
    /* --fit_order */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_spc_flat.fit_order");
    isaac_spc_flat_config.fit_order = cpl_parameter_get_int(par);
    /* --fit_size */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_spc_flat.fit_size");
    isaac_spc_flat_config.fit_size = cpl_parameter_get_int(par);
    /* --offset */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_spc_flat.offset");
    isaac_spc_flat_config.offset = cpl_parameter_get_int(par);
    /* --zone */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_spc_flat.zone"); 
    sval = cpl_parameter_get_string(par);
    if (sscanf(sval, "%d,%d,%d,%d",
                    &isaac_spc_flat_config.llx,
                    &isaac_spc_flat_config.lly,
                    &isaac_spc_flat_config.urx,
                    &isaac_spc_flat_config.ury)!=4) {
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Identify the RAW and CALIB frames in the input frameset */
    if (isaac_dfs_set_groups(framelist)) {
        cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames");
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Retrieve raw frames */
    if ((flatframes = isaac_extract_frameset(framelist,
                    ISAAC_SPC_FLAT_RAW)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot find flat frames in the input list");
        return CPL_ERROR_UNSPECIFIED;
    }

    /* The number of frames must be even */
    if (cpl_frameset_get_size(flatframes) % 2) {
        cpl_msg_error(cpl_func, "An even nb of frames expected in input");
        cpl_frameset_delete(flatframes);
        return CPL_ERROR_UNSPECIFIED;
    }
    
    /* Labelise all input frames */
    if ((labels = cpl_frameset_labelise(flatframes, isaac_spc_flat_compare, 
                &nlabels)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot labelise input frames");
        cpl_frameset_delete(flatframes);
        return CPL_ERROR_UNSPECIFIED;
    }
   
    /* Extract settings and reduce each of them */
    for (i=0; i<nlabels; i++) {
        cpl_errorstate prestate = cpl_errorstate_get();

        /* Reduce data set nb i */
        cpl_msg_info(cpl_func, "Reduce data set no %d out of %d", i+1,
                     (int)nlabels);
        cpl_msg_indent_more();
        flat_one = cpl_frameset_extract(flatframes, labels, i);
        spflat = isaac_spc_flat_reduce(flat_one);
        cpl_msg_indent_less();

        /* Save the products */
        cpl_msg_info(cpl_func, "Save the products");
        cpl_msg_indent_more();
        if (!cpl_errorstate_is_equal(prestate)) {
            cpl_image_delete(spflat);
            irplib_error_recover(prestate, "Could not reduce set %d/%d", 1+i,
                                 (int)nlabels);
        } else if (spflat == NULL) {
            cpl_msg_warning(cpl_func, "Cannot reduce set nb %d", i+1);
        } else {
            isaac_spc_flat_save(spflat, i+1, flat_one, parlist, framelist);
            cpl_image_delete(spflat);
            did_reduce = CPL_TRUE;
        }
        cpl_msg_indent_less();
        cpl_frameset_delete(flat_one);
    }
    
    /* Free and return */
    cpl_frameset_delete(flatframes);
    cpl_free(labels); 

    cpl_ensure_code(did_reduce, CPL_ERROR_ILLEGAL_INPUT);

    return cpl_error_set_where(cpl_func); /* Propagate error, if any */
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    flatframes  the frames list with flat frames
  @parma    set_nb      the setting number
  @return   0 if ok, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_image * isaac_spc_flat_reduce(cpl_frameset * flatframes)
{
    cpl_imagelist       *   diffs;
    cpl_vector          *   medians;
    double                  med, mean;
    cpl_image           *   avg_ima;
    cpl_image           *   fitted;
    int                     nima, nx, ny;
    int                     i;

    /* Test entries */
    if (flatframes == NULL) return NULL;

    /* Load input image set */
    cpl_msg_info(cpl_func, "Compute the difference images");
    if ((diffs = isaac_spc_flat_diffs(flatframes)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot create the difference images");
        return NULL;
    }
    nima = cpl_imagelist_get_size(diffs);
    nx = cpl_image_get_size_x(cpl_imagelist_get(diffs, 0));
    ny = cpl_image_get_size_y(cpl_imagelist_get(diffs, 0));

    /* Compute medians on diff images */
    cpl_msg_info(cpl_func, "Compute statistics on images");
    cpl_msg_indent_more();
    medians = cpl_vector_new(nima);

    /* Compute some stats on input images */
    for (i=0; i<nima; i++) {
        med = cpl_image_get_median_window(cpl_imagelist_get(diffs, i),
                (nx-MEDIAN_XSIZE)/2.0, (ny-MEDIAN_YSIZE)/2.0,
                (nx+MEDIAN_XSIZE)/2.0, (ny+MEDIAN_YSIZE)/2.0);
        if (cpl_vector_set(medians, i, med) != CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "Cannot compute the medians");
            cpl_vector_delete(medians);
            cpl_imagelist_delete(diffs);
            cpl_msg_indent_less();
            return NULL;
        }
    }

    /* Compute stdev and mean of the medians */
    isaac_spc_flat_config.med_avg   = cpl_vector_get_mean(medians);
    if (nima >= 2) 
        isaac_spc_flat_config.med_stdev=cpl_vector_get_stdev(medians);
    cpl_vector_delete(medians);
    cpl_msg_info(cpl_func, "Average of the medians: %g", 
            isaac_spc_flat_config.med_avg);
    cpl_msg_info(cpl_func, "Standard deviation of the medians: %g", 
            isaac_spc_flat_config.med_stdev);
    cpl_msg_indent_less();

    /* Divide by the mean in the vig in the difference image */
    cpl_msg_info(cpl_func, "Normalise the images");
    for (i=0; i<nima; i++) {

        /* Support for window readout */
        if ((nx != 1024) || (ny != 1024)) {
            mean = cpl_image_get_mean(cpl_imagelist_get(diffs, i));
        } else {
            mean = cpl_image_get_mean_window(cpl_imagelist_get(diffs, i),
                    isaac_spc_flat_config.llx, isaac_spc_flat_config.lly,
                    isaac_spc_flat_config.urx, isaac_spc_flat_config.ury);
        }
        if (cpl_image_divide_scalar(cpl_imagelist_get(diffs, i), mean) != 
                CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "Cannot normalise the image");
            cpl_imagelist_delete(diffs);
            return NULL;
        }
    }

    /* Replace by 0 the pixels whose value is <low and >high */
    for (i=0; i<nima; i++) {
        if (cpl_image_threshold(cpl_imagelist_get(diffs, i), 
                isaac_spc_flat_config.low_thresh,
                isaac_spc_flat_config.high_thresh,
                0.0, 0.0) != CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "Cannot threshold the image");
            cpl_imagelist_delete(diffs);
            return NULL;
        }
    }

    /* Average the image list to one image */
    cpl_msg_info(cpl_func, "Stack the images");
    if ((avg_ima = cpl_imagelist_collapse_create(diffs)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot collapse the image list");
        cpl_imagelist_delete(diffs);
        return NULL;
    }
    cpl_imagelist_delete(diffs);

    /* Divide the output image by the fit    */
    cpl_msg_info(cpl_func, "Divide by a fit");
    if ((fitted = isaac_spc_flat_divide_fit(avg_ima,
                    isaac_spc_flat_config.fit_order,
                    isaac_spc_flat_config.fit_size,
                    isaac_spc_flat_config.offset)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot collapse the image list");
        cpl_image_delete(avg_ima);
        return NULL;
    }
    cpl_image_delete(avg_ima);
    
    return fitted;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create the difference images
  @param    in      the input frameset
  @return   the imagelist with the difference images
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * isaac_spc_flat_diffs(cpl_frameset * in) 
{
    int                     nima;
    cpl_imagelist       *   out;
    cpl_image           *   in1;
    cpl_image           *   in2;
    cpl_frame           *   cur_frame;
    int                     i;
 
    /* Initialise */
    nima = cpl_frameset_get_size(in);

    /* Expect an even number of frames */
    if (nima % 2) {
        cpl_msg_error(cpl_func, "Odd nb of frames");
        return NULL;
    }

    /* Create the output image list */
    out = cpl_imagelist_new();

    /* Loop on all pairs */
    for (i=0; i<nima/2; i++) {
        cur_frame = cpl_frameset_get_position(in, 2*i);
        if ((in1 = cpl_image_load(cpl_frame_get_filename(cur_frame), 
                        CPL_TYPE_FLOAT, 0, 0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the image %d", 2*i);
            cpl_imagelist_delete(out);
            return NULL;
        }
        cur_frame = cpl_frameset_get_position(in, 2*i+1);
        if ((in2 = cpl_image_load(cpl_frame_get_filename(cur_frame), 
                        CPL_TYPE_FLOAT, 0, 0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the image %d", 2*i+1);
            cpl_image_delete(in1);
            cpl_imagelist_delete(out);
            return NULL;
        }
        cpl_image_subtract(in1, in2);
        cpl_image_delete(in2);
        cpl_imagelist_set(out, in1, i);
    }
    return out;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Divide an image by a fit of itself
  @param    in      the input image
  @return   the result image
 */
/*----------------------------------------------------------------------------*/
static cpl_image * isaac_spc_flat_divide_fit(
        cpl_image   *   in,
        int             order,
        int             xsize,
        int             offset) 
{
    int                     nx, ny;
    int                     xstart, xend, ystart, yend;
    cpl_image           *   collapsed;
    float               *   pcollapsed;
    cpl_image           *   extracted;
    float               *   pextracted;
    int                     nb_samples;
    cpl_matrix          *   xvec;
    cpl_vector          *   yvec;
    cpl_polynomial      *   poly_fit;
    cpl_polynomial      *   poly_2d;
    cpl_image           *   fit_image;
    cpl_image           *   out;
    cpl_size                power[2];
    int                     i;
    const cpl_size          maxdeg1d = order - 1;
    const cpl_boolean       sampsym = CPL_TRUE; /* xvec is symmetric */
    
    /* Test entries */
    if (in == NULL) return NULL;

    /* Initialise */
    nx = cpl_image_get_size_x(in);
    ny = cpl_image_get_size_y(in);
    
    /* Determine the zone to extract */
    xstart = (int)((nx - xsize)/2) + 1;
    xend = xstart + xsize - 1;
    if ((xstart<1) || (xend>nx)) {
        cpl_msg_error(cpl_func, "bad X size specified");
        return NULL;
    }

    /* Extract the central zone and collapse it */
    if ((collapsed = cpl_image_collapse_window_create(in, xstart, 1, xend, ny, 
                    1)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot collpase a part of the image");
        return NULL;
    }
    pcollapsed = cpl_image_get_data_float(collapsed);

    /* Find the 'valid' zone in the 1D image */
    ystart = 1;
    while ((fabs(pcollapsed[ystart-1]) < 1e-4) && (ystart < nx)) ystart++;
    ystart += offset;

    yend = ny;
    while ((fabs(pcollapsed[yend-1]) <1e-4) && (yend > 1)) yend--;
    yend -= offset;

    if (ystart > yend) {
        cpl_msg_error(cpl_func, "invalid coordinates of the zone to extract");
        cpl_image_delete(collapsed);
        return NULL;
    }

    /* Extract the 1D signal to fit */
    if ((extracted = cpl_image_extract(collapsed, 1, ystart, 1, yend))==NULL) {
        cpl_msg_error(cpl_func, "cannot extract 1D image");
        cpl_image_delete(collapsed);
        return NULL;
    }
    cpl_image_delete(collapsed);
    pextracted = cpl_image_get_data_float(extracted);

    /* Fit the polynomial */
    nb_samples = cpl_image_get_size_y(extracted);
    xvec = cpl_matrix_new(1, nb_samples);
    yvec = cpl_vector_new(nb_samples);
    for (i=0; i<nb_samples; i++) {
        cpl_matrix_set(xvec, 0, i, (double)(ystart + i));
        cpl_vector_set(yvec,    i, (double)(pextracted[i] / xsize));
    }
    cpl_image_delete(extracted);
    poly_fit = cpl_polynomial_new(1);
    if (cpl_polynomial_fit(poly_fit, xvec, &sampsym, yvec, NULL, CPL_FALSE,
                           NULL, &maxdeg1d)) {
        cpl_msg_error(cpl_func, "cannot fit the 1D signal");
        cpl_matrix_delete(xvec);
        cpl_vector_delete(yvec);
        cpl_polynomial_delete(poly_fit);
        return NULL;
    }
    cpl_matrix_delete(xvec);
    cpl_vector_delete(yvec);
    
    /* The polynomial for image generation must be 2d */
    poly_2d = cpl_polynomial_new(2);
    power[0] = 0; power[1] = 0;
    cpl_polynomial_set_coeff(poly_2d, power, 
            cpl_polynomial_get_coeff(poly_fit, power + 1));
    power[0] = 0; power[1] = 1;
    cpl_polynomial_set_coeff(poly_2d, power, 
            cpl_polynomial_get_coeff(poly_fit, power + 1));
    power[0] = 0; power[1] = 2;
    cpl_polynomial_set_coeff(poly_2d, power, 
            cpl_polynomial_get_coeff(poly_fit, power + 1));
    cpl_polynomial_delete(poly_fit);
    
    /* Create the fit image */
    fit_image = cpl_image_new(nx, ny, CPL_TYPE_FLOAT);
    cpl_image_fill_polynomial(fit_image, poly_2d, 1.0, 1.0, 1.0, 1.0);
    cpl_polynomial_delete(poly_2d);
    
    /* Divide the input image by the polynomial image */
    if ((out = cpl_image_divide_create(in, fit_image)) == NULL) {
        cpl_msg_error(cpl_func, "cannot divide the images");
        cpl_image_delete(fit_image);
        return NULL;
    }
    cpl_image_delete(fit_image);

    return out;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the lampflat recipe products on disk
  @param    flat    the flat image produced
  @param    set_nb  current setting number
  @param    set     the frames from the current setting
  @param    parlist     the input list of parameters
  @param    set_tot the input frame set
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_flat_save(
        cpl_image           *   flat,
        int                     set_nb,
        cpl_frameset        *   set,
        cpl_parameterlist   *   parlist,
        cpl_frameset        *   set_tot)
{
    cpl_propertylist    *   plist;
    cpl_propertylist    *   qclist;
    cpl_propertylist    *   paflist;
    const cpl_frame           *   ref_frame;
    const char          *   sval;
    int                     arm;
    const char          *   procat;
    char                *   filename;

    /* Get the QC params in qclist */
    qclist = cpl_propertylist_new();
 
    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }
    /* Test the status */
    if (cpl_error_get_code()) {
        cpl_propertylist_delete(qclist);
        cpl_propertylist_delete(plist);
        return CPL_ERROR_UNSPECIFIED;
    }
   /* Get the arm used */
    sval = isaac_pfits_get_arm(plist);
    if (sval==NULL) {
        cpl_msg_error(cpl_func, "Cannot get the arm used");
        cpl_propertylist_delete(plist);
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }
    if (sval[0] == 'S') arm = 1;
    else if (sval[0] == 'L') arm = 2;
    else {
        cpl_msg_error(cpl_func, "Unsupported arm");
        cpl_propertylist_delete(plist);
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }
    sval = isaac_pfits_get_filter(plist);
    if (cpl_error_get_code()) cpl_error_reset();
    else cpl_propertylist_append_string(qclist, "ESO QC FILTER OBS", sval);
    cpl_propertylist_delete(plist);
    cpl_propertylist_append_double(qclist, "ESO QC SPECFLAT NCOUNTS",
            isaac_spc_flat_config.med_avg);
    cpl_propertylist_append_double(qclist, "ESO QC SPECFLAT STDEV",
            isaac_spc_flat_config.med_stdev);

    /* Write the flat image */
    procat = arm == 1 ? ISAAC_SPC_FLAT_SW_RES : ISAAC_SPC_FLAT_LW_RES;
    filename = cpl_sprintf("isaac_spc_flat_set%02d.fits", set_nb);
    irplib_dfs_save_image(set_tot,
            parlist,
            set,
            flat,
            CPL_BPP_IEEE_FLOAT,
            "isaac_spc_flat",
            procat,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            filename);
    cpl_free(filename);

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);

    /* Get FITS header from reference file */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }
    
    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist, 
        "^(ARCFILE|MJD-OBS|INSTRUME|ESO TPL ID|ESO TPL NEXP|ESO DPR CATG|"
        "ESO DPR TECH|ESO DPR TYPE|DATE-OBS|ESO INS GRAT NAME|"
        "ESO INS GRAT WLEN|ESO INS OPTI1 ID|ESO DET DIT|ESO INS LAMP3 SET)$",0);
    cpl_propertylist_delete(plist);
    
    /* Copy the QC in paflist */
    cpl_propertylist_copy_property_regexp(paflist, qclist, ".", 0);
    cpl_propertylist_delete(qclist);

    /* PRO.CATG */
    cpl_propertylist_update_string(paflist, CPL_DFS_PRO_CATG, procat);

    /* Save the PAF file */
    filename = cpl_sprintf("isaac_spc_flat_set%02d.paf", set_nb);
    cpl_dfs_save_paf("ISAAC",
            "isaac_spc_flat",
            paflist,
            filename);
    cpl_free(filename);
    cpl_propertylist_delete(paflist);
    return  0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Comparison function to identify different settings
  @param    frame1  first frame 
  @param    frame2  second frame 
  @return   0 if frame1!=frame2, 1 if frame1==frame2, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_flat_compare(
        const cpl_frame *   frame1, 
        const cpl_frame *   frame2) 
{
    int                     comparison;
    cpl_propertylist    *   plist1;
    cpl_propertylist    *   plist2;
    const char          *   sval1,
                        *   sval2;
    double                  dval1, dval2;

    /* Test entries */
    if (frame1==NULL || frame2==NULL) return CPL_ERROR_UNSPECIFIED;

    /* Get property lists */
    if ((plist1=cpl_propertylist_load(cpl_frame_get_filename(frame1),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        return CPL_ERROR_UNSPECIFIED;
    }
    if ((plist2=cpl_propertylist_load(cpl_frame_get_filename(frame2),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(plist1);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Test status */
    if (cpl_error_get_code()) {
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }

    comparison = 1;

    /* Compare the slit used */
    sval1 = isaac_pfits_get_opti1_id(plist1);
    sval2 = isaac_pfits_get_opti1_id(plist2);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get the slit used");
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }
    if (strcmp(sval1, sval2)) comparison = 0;

    /* Compare the resolution */
    sval1 = isaac_pfits_get_resolution(plist1);
    sval2 = isaac_pfits_get_resolution(plist2);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get the resolution");
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }
    if (strcmp(sval1, sval2)) comparison = 0;

    /* Compare the central wavelength */
    dval1 = isaac_pfits_get_wlen(plist1);
    dval2 = isaac_pfits_get_wlen(plist2);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get the central wavelength");
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }
    if (fabs(dval1-dval2) > 1e-4) comparison = 0;

    cpl_propertylist_delete(plist1);
    cpl_propertylist_delete(plist2);
    return comparison;
}


