/* $Id: isaac_spc_wlcalib.c,v 1.26 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.26 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <cpl.h>

#include "irplib_utils.h"

#include "isaac_utils.h"
#include "isaac_wavelength.h"
#include "isaac_physicalmodel.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int isaac_spc_wlcalib_create(cpl_plugin *);
static int isaac_spc_wlcalib_exec(cpl_plugin *);
static int isaac_spc_wlcalib_destroy(cpl_plugin *);
static int isaac_spc_wlcalib(cpl_parameterlist *, cpl_frameset *);
static cpl_table * isaac_spc_wlcalib_compute(const char *, const char *, 
        const char *, const char *);
static int isaac_spc_wlcalib_save(const cpl_table *, int, cpl_parameterlist *, 
        cpl_frameset *);
 
/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int         rej_left;
    int         rej_right;
    int         rej_bottom;
    int         rej_top;
    int         max_offset;
    int         lines_table;
    /* Outputs */
    double      disprel_cc;
    int         disprel_clines;
    int         disprel_dlines;
    double      disprel_rms;
} isaac_spc_wlcalib_config;

static char isaac_spc_wlcalib_description[] = 
"isaac_spc_wlcalib -- ISAAC wavelength calibration recipe\n" 
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"oh-cat.fits "ISAAC_CALPRO_OH_CAT" or\n"
"xe-cat.fits "ISAAC_CALPRO_XE_CAT" or\n"
"ar-cat.fits "ISAAC_CALPRO_AR_CAT"\n";

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe));
    cpl_plugin  *   plugin = &recipe->interface;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    ISAAC_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "isaac_spc_wlcalib",
                    "Spectro wavelength calibration",
                    isaac_spc_wlcalib_description,
                    "Lars Lundin",
                    PACKAGE_BUGREPORT,
                    isaac_get_license(),
                    isaac_spc_wlcalib_create,
                    isaac_spc_wlcalib_exec,
                    isaac_spc_wlcalib_destroy);

    cpl_pluginlist_append(list, plugin);
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_wlcalib_create(cpl_plugin * plugin)
{
    cpl_recipe      * recipe;
    cpl_parameter   * p;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();

    /* Fill the parameters list */
    /* --rejected */
    p = cpl_parameter_new_value("isaac.isaac_spc_wlcalib.rejected",
            CPL_TYPE_STRING, "left right bottom top rejections", 
            "isaac.isaac_spc_wlcalib", "-1 -1 100 100");
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "rej");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --max_offset */
    p = cpl_parameter_new_value("isaac.isaac_spc_wlcalib.max_offset",
            CPL_TYPE_INT, "maximum offset from the physical model "
            "allowed in pixels", "isaac.isaac_spc_wlcalib", 50);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "max_offset");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --lines */
    p = cpl_parameter_new_value("isaac.isaac_spc_wlcalib.lines", CPL_TYPE_INT, 
            "1 for oh, 2 for xe, 3 for ar, 4 for xe+ar",
            "isaac.isaac_spc_wlcalib", -1);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "lines");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_wlcalib_exec(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    return isaac_spc_wlcalib(recipe->parameters, recipe->frames);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_wlcalib_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    cpl_parameterlist_delete(recipe->parameters);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_wlcalib(
        cpl_parameterlist   *   parlist, 
        cpl_frameset        *   framelist)
{
    cpl_parameter       *   par;
    cpl_frame           *   cur_frame;
    const char          *   cur_fname;
    const char          *   sval;
    const char          *   oh;
    const char          *   xe;
    const char          *   ar;
    cpl_table           *   tab;
    
    /* Initialise */
    par = NULL;

    /* Retrieve input parameters */
    /* Rejection parameters */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_spc_wlcalib.rejected");
    sval = cpl_parameter_get_string(par);
    if (sscanf(sval, "%d %d %d %d",
                    &isaac_spc_wlcalib_config.rej_left,
                    &isaac_spc_wlcalib_config.rej_right,
                    &isaac_spc_wlcalib_config.rej_bottom,
                    &isaac_spc_wlcalib_config.rej_top) != 4) {
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Max offset in pixels */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_spc_wlcalib.max_offset");
    isaac_spc_wlcalib_config.max_offset = cpl_parameter_get_int(par);

    /* --lines */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_spc_wlcalib.lines");
    isaac_spc_wlcalib_config.lines_table=cpl_parameter_get_int(par);

    /* Check lines */
    if (isaac_spc_wlcalib_config.lines_table < 1) {
        cpl_msg_error(cpl_func, "Please specify the catalog");
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Retrieve calibration data */
    oh = isaac_extract_filename(framelist, ISAAC_CALPRO_OH_CAT);
    xe = isaac_extract_filename(framelist, ISAAC_CALPRO_XE_CAT);
    ar = isaac_extract_filename(framelist, ISAAC_CALPRO_AR_CAT);
    
    /* Reduce the images */
    cur_frame = cpl_frameset_get_position(framelist, 0);
    cur_fname = cpl_frame_get_filename(cur_frame);
    cpl_msg_info(cpl_func, "Reduce %s", cur_fname);
    cpl_msg_indent_more();

    /* Apply the wl calib */
    if ((tab = isaac_spc_wlcalib_compute(cur_fname, oh, ar, xe)) == NULL) {
        cpl_msg_warning(cpl_func, "cannot reduce");
    } else {
        /* Save the file */
        if (isaac_spc_wlcalib_save(tab, 1, parlist, framelist) != 0) {
             cpl_msg_warning(cpl_func, "cannot save");
        }
        cpl_table_delete(tab);
    }
    cpl_msg_indent_less();
    
    /* Free and return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Determine the distortion, correct, and wl calibrate
  @param    in_name     input file name
  @param    oh          the oh lines catalog
  @param    ar          the ar lines catalog
  @param    xe          the xe lines catalog
  @return   the result table with 1 column with the 1D polynomial describing 
            the dispersion
 */
/*----------------------------------------------------------------------------*/
static cpl_table * isaac_spc_wlcalib_compute(
        const char          *   in_name,
        const char          *   oh,
        const char          *   ar,
        const char          *   xe)
{
    double                  slit_width;
    int                     order;
    double              *   phdisprel;
    cpl_image           *   in;
    computed_disprel    *   disprel;
    cpl_table           *   out_tab;
    const char          *   lines_tab;

    /* Compute the slit_width */
    if ((slit_width = isaac_get_slitwidth(in_name)) == -1) {
        cpl_msg_error(cpl_func, "cannot get the slit width");
        return NULL;
    }

    /* Get the order */
    if ((order = isaac_find_order(in_name)) == -1) {
        cpl_msg_error(cpl_func, "cannot get the order");
        return NULL;
    }

    /* First estimation using a physical model */
    cpl_msg_info(cpl_func, "Compute the physical model");
    cpl_msg_indent_more();
    if ((phdisprel = isaac_get_disprel_estimate(in_name, 3)) == NULL) {
        cpl_msg_error(cpl_func, "cannot estimate the dispersion relation");
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_msg_info(cpl_func, "f(x)=%g + %g*x + %g*x^2 + %g*x^3",
            phdisprel[0], phdisprel[1], phdisprel[2], phdisprel[3]);
    cpl_msg_indent_less();

    /* Load the image */
    if ((in = cpl_image_load(in_name, CPL_TYPE_FLOAT, 0, 0)) == NULL) {
        cpl_msg_error(cpl_func, "cannot load the image");
        cpl_free(phdisprel);
        return NULL;
    }
    
    /* Compute the dispersion relation */
    switch (isaac_spc_wlcalib_config.lines_table) {
        case 1: lines_tab = "oh"; break;
        case 2: lines_tab = "Xe"; break;
        case 3: lines_tab = "Ar"; break;
        case 4: lines_tab = "Xe+Ar"; break;
        default: 
                cpl_image_delete(in);
                return NULL;
    }
    cpl_msg_info(cpl_func, "Compute the dispersion relation");
    cpl_msg_indent_more();
    if ((disprel = spectro_compute_disprel(in, 
                    isaac_spc_wlcalib_config.rej_bottom,
                    isaac_spc_wlcalib_config.rej_top,
                    isaac_spc_wlcalib_config.rej_left,
                    isaac_spc_wlcalib_config.rej_right,
                    isaac_spc_wlcalib_config.max_offset,
                    isaac_has_thermal(in_name) > 0,
                    lines_tab, oh, ar, xe, slit_width, order, 
                    (int)(cpl_msg_get_level() == CPL_MSG_DEBUG), 
                    phdisprel)) == NULL) {
        cpl_msg_error(cpl_func, "cannot compute the dispersion relation");
        cpl_image_delete(in);
        cpl_free(phdisprel);
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_image_delete(in);
    cpl_free(phdisprel);
    
    cpl_msg_info(cpl_func, "Cross correlation factor: %g", disprel->cc);
    cpl_msg_info(cpl_func, "f(x)=%g + %g*x + %g*x^2 + %g*x^3",
            disprel->poly[0], disprel->poly[1], disprel->poly[2],
            disprel->poly[3]);
    cpl_msg_indent_less();
    
    /* Create the output table and store the relevant information */
    out_tab = cpl_table_new(4);
    
    /* Fill dispersion relation */
    cpl_table_new_column(out_tab, "WL_coefficients", CPL_TYPE_DOUBLE);
    cpl_table_set_double(out_tab, "WL_coefficients", 0, disprel->poly[0]);
    cpl_table_set_double(out_tab, "WL_coefficients", 1, disprel->poly[1]);
    cpl_table_set_double(out_tab, "WL_coefficients", 2, disprel->poly[2]);
    cpl_table_set_double(out_tab, "WL_coefficients", 3, disprel->poly[3]);
    isaac_spc_wlcalib_config.disprel_cc     = disprel->cc;
    isaac_spc_wlcalib_config.disprel_clines = disprel->clines;
    isaac_spc_wlcalib_config.disprel_dlines = disprel->dlines;
    isaac_spc_wlcalib_config.disprel_rms    = disprel->rms;

    /* Free and return */
    if (disprel->poly != NULL) cpl_free(disprel->poly);
    cpl_free(disprel);
    return out_tab;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the recipe products on disk
  @param    tab         table to save
  @param    nb          the frame number
  @param    parlist     the input list of parameters
  @param    set         the input frame set
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_wlcalib_save(
        const cpl_table     *   tab,
        int                     nb,
        cpl_parameterlist   *   parlist,
        cpl_frameset        *   set)
{
    cpl_propertylist    *   plist;
    cpl_propertylist    *   qclist;
    cpl_propertylist    *   paflist;
    const cpl_frame           *   ref_frame;
    const char          *   sval;
    char                *   filename;

    /* Get the QC params in qclist */
    qclist = cpl_propertylist_new();

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }
    /* Test the status */
    if (cpl_error_get_code()) {
        cpl_propertylist_delete(qclist);
        cpl_propertylist_delete(plist);
        return CPL_ERROR_UNSPECIFIED;
    }
    sval = isaac_pfits_get_filter(plist);
    if (cpl_error_get_code()) cpl_error_reset();
    else cpl_propertylist_append_string(plist, "ESO QC FILTER OBS", sval);
    cpl_propertylist_delete(plist);
    cpl_propertylist_append_int(plist, "ESO QC LAMP", 
            isaac_spc_wlcalib_config.lines_table);
    cpl_propertylist_append_double(plist, "ESO QC DISP XCORR", 
            isaac_spc_wlcalib_config.disprel_cc);
    cpl_propertylist_append_int(plist, "ESO QC DISP NUMCAT", 
            isaac_spc_wlcalib_config.disprel_clines);
    cpl_propertylist_append_int(plist, "ESO QC DISP NUMMATCH", 
            isaac_spc_wlcalib_config.disprel_dlines);
    cpl_propertylist_append_double(plist, "ESO QC DISP STDEV", 
            isaac_spc_wlcalib_config.disprel_rms);
    cpl_propertylist_append_double(plist, "ESO QC DISPCO1", 
            cpl_table_get_double(tab, "WL_coefficients", 0, NULL));
    cpl_propertylist_append_double(plist, "ESO QC DISPCO2", 
            cpl_table_get_double(tab, "WL_coefficients", 1, NULL));
    cpl_propertylist_append_double(plist, "ESO QC DISPCO3", 
            cpl_table_get_double(tab, "WL_coefficients", 2, NULL));
    cpl_propertylist_append_double(plist, "ESO QC DISPCO4", 
            cpl_table_get_double(tab, "WL_coefficients", 3, NULL));
    cpl_propertylist_append_double(plist, "ESO QC WLEN",
        (double)(cpl_table_get_double(tab, "WL_coefficients", 0, NULL) +
             512*cpl_table_get_double(tab, "WL_coefficients", 1, NULL) +
         512*512*cpl_table_get_double(tab, "WL_coefficients", 2, NULL) +
     512*512*512*cpl_table_get_double(tab, "WL_coefficients", 3, NULL)));
 
    /* Write the table  */
    filename = cpl_sprintf("isaac_spc_wlcalib_%d.fits", nb);
    irplib_dfs_save_table(set,
            parlist,
            set,
            tab,
            NULL,
            "isaac_spc_wlcalib",
            ISAAC_SPC_WLCALIB_RES,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            filename);
    cpl_free(filename);
    
    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);

    /* Get FITS header from reference file */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }
    
    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist, 
        "^(ARCFILE|MJD-OBS|INSTRUME|ESO TPL ID|ESO TPL NEXP|ESO DPR CATG|"
        "ESO DPR TECH|ESO DPR TYPE|DATE-OBS|ESO INS GRAT NAME|"
        "ESO INS GRAT WLEN|ESO INS GRAT ORDER|ESO INS MODE|"
        "ESO INS OPTI1 ID)$", 0);
    cpl_propertylist_delete(plist);

    /* Copy the QC in paflist */
    cpl_propertylist_copy_property_regexp(paflist, qclist, ".", 0);
    cpl_propertylist_delete(qclist);

    /* PRO.CATG */
    cpl_propertylist_update_string(paflist, CPL_DFS_PRO_CATG,
                                   ISAAC_SPC_WLCALIB_RES);

    /* Save the PAF file */
    filename = cpl_sprintf("isaac_spc_wlcalib_%d.paf", nb);
    cpl_dfs_save_paf("ISAAC",
            "isaac_spc_wlcalib",
            paflist,
            filename);
    cpl_free(filename);
    cpl_propertylist_delete(paflist);
    return  0;
}

