/* $Id: isaac_util_seds.c,v 1.15 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.15 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "irplib_utils.h"

#include "isaac_utils.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"

#include <string.h>
#include <math.h>

/*-----------------------------------------------------------------------------
                                Defines
 -----------------------------------------------------------------------------*/

#define RECIPE_NAME "isaac_util_seds"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

cpl_recipe_define(isaac_util_seds, ISAAC_BINARY_VERSION,
                  "Lars Lundin", PACKAGE_BUGREPORT, "2008", 
                  "ISAAC SEDS table creation",
                  RECIPE_NAME " -- ISAAC SEDS table creation.\n"
                  "The files listed in the Set Of Frames (sof-file) " 
                  "must be tagged:\n"
                  "ASCII-SEDS.txt " ISAAC_UTIL_SEDS_RAW "\n"
                  "The first three characters of the input file names must be "
                  "unique. Lines starting with a '#' are ignored. The other "
                  "lines must have (at least) two columns, the first with a "
                  "wavelength, the second with an SED. All files must have "
                  "an identical number of non-comment rows and all files are "
                  "assumed to contain identical sequences of wavelengths.\n");

static
cpl_error_code isaac_util_seds_save(cpl_frameset *, const cpl_table *,
                                    const cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Fill the recipe parameterlist
  @param    self  The parameterlist
  @return   0 iff everything is ok

 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code isaac_util_seds_fill_parameterlist(cpl_parameterlist * self) {

    return self != NULL ? CPL_ERROR_NONE
        : cpl_error_set(cpl_func, CPL_ERROR_NULL_INPUT);
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    The recipe data reduction part is implemented here 
  @param    framelist   the frames list
  @param    parlist     the parameters list
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_util_seds(cpl_frameset            * framelist,
                           const cpl_parameterlist * parlist)
{
    cpl_frameset * rawframes = NULL;
    cpl_table    * out       = NULL;
    cpl_bivector * seds      = NULL;
    cpl_vector   * sedsx     = NULL;
    cpl_vector   * sedsx1    = NULL;
    cpl_vector   * sedsy     = NULL;
    int            nframes   = 0;
    int            nlines    = 0;
    int            i;
    
    /* Identify the RAW and CALIB frames in the input frameset */
    skip_if (isaac_dfs_set_groups(framelist));

    /* Retrieve raw frames */
    rawframes = isaac_extract_frameset(framelist, ISAAC_UTIL_SEDS_RAW);
    irplib_ensure(rawframes != NULL, CPL_ERROR_DATA_NOT_FOUND,
                  "Could not find raw frames in the input list");

    nframes = cpl_frameset_get_size(rawframes);
    irplib_ensure(nframes > 0, CPL_ERROR_DATA_NOT_FOUND,
                  "Could not find raw frames in the input list");
 
    /* Loop over ASCII files */
    for (i = 0; i < nframes; i++) {
        const cpl_frame * rawframe = cpl_frameset_get_position_const(rawframes, i);

        /* Create the 3-letter label name from the catalogue name */
        char name[4];
        const char * rawfile = cpl_frame_get_filename(rawframe);
        /* Skip path, if any */
        const char * cat_name = strrchr(rawfile, '/');

        cat_name = cat_name ? 1+cat_name : rawfile;
        bug_if(cat_name == NULL);
        (void)strncpy(name, cat_name, 3);
        name[3] = '\0';

        /* Load ASCII file */
        seds = cpl_bivector_read(cpl_frame_get_filename(rawframe));
        irplib_ensure(seds != NULL, cpl_error_get_code(),
                      "Could not load SED no. %d of %d", i+1, nframes);
        sedsx = cpl_bivector_get_x(seds);
        sedsy = cpl_bivector_get_y(seds);
        cpl_bivector_unwrap_vectors(seds);
        seds = NULL;
        
        if (i == 0) {
            nlines = cpl_vector_get_size(sedsx);

            /* Create the table */
            out = cpl_table_new(nlines);
            /* - and the wavelength column */
            bug_if(cpl_table_wrap_double(out, cpl_vector_get_data(sedsx),
                                         "Wavelength"));
            sedsx1 = sedsx;
            sedsx = NULL;
        } else {
            /* Verify the consistency of length */
            const int ilines = cpl_vector_get_size(sedsx);
            double dmin, dmax;

            irplib_ensure(ilines == nlines, CPL_ERROR_INCOMPATIBLE_INPUT,
                          "Number of lines in SED no. %d and %d differ: "
                          "%d <=> %d", 1, i+1, nlines, ilines);

            /* FIXME: Support different wavelengths ? */
            bug_if(cpl_vector_subtract(sedsx, sedsx1));

            dmin = cpl_vector_get_min(sedsx);
            dmax = cpl_vector_get_max(sedsx);

            if (dmin != 0.0 || dmax != 0.0) {
                cpl_msg_warning(cpl_func, "Wavelengths in SED no. %d differ "
                                "from those in no. 1 (min, max): %g %g",
                                i+1, dmin, dmax);
            }

            cpl_vector_delete(sedsx);
            sedsx = NULL;
        }

        /* Create the column */
        skip_if(cpl_table_wrap_double(out, cpl_vector_get_data(sedsy), name));
        (void)cpl_vector_unwrap(sedsy);
        sedsy = NULL;
    }

    /* Save the table */
    skip_if (isaac_util_seds_save(framelist, out, parlist));

    end_skip;
    
    cpl_bivector_unwrap_vectors(seds);
    (void)cpl_vector_unwrap(sedsx1);
    cpl_vector_delete(sedsx);
    cpl_vector_delete(sedsy);
    cpl_frameset_delete(rawframes);
    cpl_table_delete(out);

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Save the SEDs table
  @param    sed     the table
  @param    parlist the input list of parameters
  @param    set     the input frame set
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code isaac_util_seds_save(cpl_frameset            * set,
                                    const cpl_table         * sed,
                                    const cpl_parameterlist * parlist)
{
    cpl_propertylist * plist = cpl_propertylist_new();

    bug_if(cpl_propertylist_append_string(plist, "INSTRUME", "ISAAC"));
 
    /* Write the table */
    skip_if(irplib_dfs_save_table(set,
                                  parlist,
                                  set,
                                  sed,
                                  NULL,
                                  RECIPE_NAME,
                                  ISAAC_UTIL_SEDS_RES,
                                  plist,
                                  NULL,
                                  PACKAGE "/" PACKAGE_VERSION,
                                  RECIPE_NAME CPL_DFS_FITS));

    end_skip;

    cpl_propertylist_delete(plist);

    return cpl_error_get_code();
}


